import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import torch
from typing import Dict, Any, Union, List, Tuple
from matplotlib.colors import LinearSegmentedColormap

class PlotSpec:
    def __init__(self, plot_type: str, data_keys: Union[str, List[str]], title: str = None, **kwargs):
        self.plot_type = plot_type
        self.data_keys = [data_keys] if isinstance(data_keys, str) else data_keys
        self.title = title
        self.kwargs = kwargs

class MatplotlibWrapper:
        
    def setup(self):
        plt.style.use('seaborn-v0_8-pastel')
        sns.set_style("whitegrid") # whitegrid, darkgrid, dark, white, ticks
        # print(plt.rcParams.keys())
        plt.rcParams['figure.dpi'] = 100
        plt.rcParams['font.family'] = 'monospace'
        plt.rcParams['font.monospace'] = ['Roboto Mono']
        # plt.rcParams['font.usetex'] = True
        plt.rcParams['font.size'] = 8
        plt.rcParams['axes.labelsize'] = 10
        plt.rcParams['axes.titlesize'] = 12
        plt.rcParams['xtick.labelsize'] = 6
        plt.rcParams['ytick.labelsize'] = 6
        plt.rcParams['legend.fontsize'] = 6
        plt.rcParams['figure.titlesize'] = 14
        self.colors = ['#FAF3DD','#C8D5B9','#8FC0A9']
        self.cmap = LinearSegmentedColormap.from_list("custom", self.colors, N=256)

    def plot(self, plot_type, cache, data_keys, title, **kwargs):
        self.setup(self) # explicitly we have to do this.
        
        method_name = f"plot_{plot_type}"
        if not hasattr(self, method_name):
            raise ValueError(f"Unknown plot type: {plot_type}")

        return getattr(self, method_name)(self, cache=cache, data_keys=data_keys, title=title, **kwargs)
    
    
    def plot_attention(self, cache, data_keys, title, **kwargs):
        attn_data = cache[data_keys[0]]
        fig, ax = plt.subplots()

        sns.heatmap(attn_data[0], ax=ax, cmap=self.cmap)
        ax.set_title(title or f"Attention Patterns : {data_keys[0]}")
        ax.set_xlabel("Key Position")
        ax.set_ylabel("Query Position")
        plt.tight_layout()
        return fig


    def plot_neuron_activation(self, cache, data_keys ,title: str = None, **kwargs) -> plt.Figure:
        # print(cache.keys())
        activations = cache[data_keys[0]].mean(dim=0)
        
        fig, ax = plt.subplots()
        sns.violinplot(data=activations.t(), ax=ax, cmap = self.cmap)
        ax.set_title(title or f"Neuron Activations L {data_keys[0]}")
        ax.set_xlabel("Neuron Index")
        ax.set_ylabel("Activation")
        plt.tight_layout()

        return fig

    def plot_attention_stats(self, data_keys, title: str = None, **kwargs) -> plt.Figure:
        # attn_data = self._ensure_tensor(data["attention"])
        # path = data.get("path")
        # tokens = data.get("tokens", None)

        # if path is None:
        src_layer, src_head = kwargs.pop("src_layer", 0), kwargs.pop("src_head", 0)
        dst_layer, dst_head = kwargs.pop("dst_layer", 1), kwargs.pop("dst_head", 0)
        path = [(src_layer, src_head, dst_layer, dst_head)]

        # Example matplotlib plotting (replace with your specific plotting logic)
        fig, ax = plt.subplots()
        # Plot attention flow
        ax.text(0.5, 0.5, "Attention Flow", fontsize=12, ha='center')
        ax.axis('off')
        fig.tight_layout()

        return fig

    def plot_induction_heads(self, cache, data_keys, title: str = None, **kwargs) -> plt.Figure:
        attn_patterns = cache[data_keys[0]]
        num_heads = attn_patterns.shape[0]

        fig, axs = plt.subplots(1,num_heads, figsize=(4*num_heads, 4))

        if num_heads > 1:
            for head in range(num_heads):
                sns.heatmap(attn_patterns[head :, ], ax=axs[head], cmap=self.cmap)
                axs[head].set_title(f"Head {head}")
                axs[head].set_xlabel("Key Position")

                if head == 0:
                    axs[head].set_ylabel("Query Position")
                else:
                    axs[head].set_ylabel("")
        else:
            sns.heatmap(attn_patterns[0], ax=axs, cmap=self.cmap)
            axs.set_xlabel("Key Position")
            axs.set_ylabel("Query Position")
            axs.set_title(f"Head 0")
        
        fig.suptitle(title or f"Induction Heads : {data_keys[0]}")
        plt.tight_layout()
        return fig
    

    def plot_neuron_activation_trajectories(self, cache, data_keys, title: str = None, **kwargs) -> plt.Figure:
        activations = cache[data_keys[0]]
        neuron_magnitudes = activations.abs().mean(dim=0)
        top_k_neurons = torch.topk(neuron_magnitudes, k=5).indices

        fig, ax = plt.subplots()
        for neuron in top_k_neurons:
            ax.plot(activations[:, neuron], label=f"Neuron {neuron}")
        
        ax.set_title(title or f"Neuron Activation Trajectories : {data_keys[0]}")
        ax.set_xlabel("Sequence Position")
        ax.set_ylabel("Activation")
        ax.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
        plt.tight_layout()
        return fig

    def plot_qk_dot_product(self, cache, data_keys, title, **kwargs):
        # print(cache.keys())

        q = cache[data_keys[0]]
        k = cache[data_keys[1]]
        
        qk = q @ k.transpose(-1, -2)

        fig, ax = plt.subplots()
        sns.heatmap(qk, ax=ax, cmap=self.cmap, center=0)
        ax.set_title(title or "Average QK Dot Products")
        ax.set_xlabel("Key Position")
        ax.set_ylabel("Query Position")
        plt.tight_layout()

        return fig


    def plot_residual_stream(self, cache, data_keys, title: str = None, **kwargs) -> plt.Figure:
        res_keys = sorted([k for k in cache.keys() if "hook_resid_post" in k])
        res_norms = [torch.norm(cache[k]).item() for k in res_keys]

        fig, ax = plt.subplots()
        fig.suptitle(title or "Residual Stream Norm Progression")

        sns.lineplot(x=range(len(res_keys)), y=res_norms, ax=ax, markers='o')
        
        ax.set_xticks(range(len(res_keys)))
        ax.set_xticklabels([f"L{i}" for i in range(len(res_keys))], ha='right')
        ax.set_xlabel("Layer")
        ax.set_ylabel("Residual Norm")
        
        plt.tight_layout()
        return fig
    
    def plot_layer_norms(self, cache, data_keys, title, **kwargs):
        ln_keys = sorted([k for k in cache.keys() if "ln" in k]) # have to see how to use model_attrs here.
        ln_stats = [torch.norm(cache[k]).item() for k in ln_keys]

        fig, ax = plt.subplots()
        ax.plot(ln_stats, marker='o')
        ax.set_xticks(range(len(ln_keys)))
        ax.set_xticklabels(ln_keys, rotation=45, ha='right')
        ax.set_title(title or "Layer Norm Statistics")
        ax.set_ylabel("Norm")
        plt.tight_layout()

        return fig


# Update: Everything works.
import torch
import torch.nn as nn
from .base_interpret import BaseInterpretabilityTool

class SparsityAnalyzer(BaseInterpretabilityTool):
    """Analyzes the sparsity of the network."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model

    # Works(Tested with GPT2)
    def compute_activation_sparsity(self, id_list, layer):
        """Computes the sparsity of the activations of the layer."""
        total_neurons = 0
        active_neurons = 0
        for input_ids in id_list:
            _, cache = self.model(input_ids)
            acts = cache[layer].squeeze() > 0
            total_neurons += acts.numel()
            active_neurons += acts.sum().item()
        return 1 - (active_neurons / total_neurons)

    # Works.
    def prune_network(self, sparsity_threshold=0.9):
        """Prunes the network based on the sparsity threshold."""
        for name, module in self.model.model.named_modules():
            if isinstance(module, nn.LayerNorm):  # I don't think this is making sense to me.
                W = module.weight.data
                # print(self.model.model_attrs.get_lin_ff_type())
                if self.model.model_attrs.get_lin_ff_type() in name:  # Don't prune the unembedding matrix. Check with our notation.
                    continue
                row_norms = torch.norm(W, dim=0)
                mask = row_norms > torch.quantile(row_norms, sparsity_threshold)
                module.weight.data = W * mask.unsqueeze(0)

    #Works.
    def find_polysemantic_neurons(self, threshold=1):
        """Finds the polysemantic neurons in the network."""

        # By deafualt, we are testing for MLP layer.
        # print(self.model.model.state_dict().keys())
        # input()
        W = self.model.model_attrs.get_mlp_out(0).weight # Let's see if it makes sense :(
        # W = self.model.model.state_dict()[layer]
        U, S, V = torch.svd(W)
        S_norm = S / S.sum()
        polysemantic = []
        for i in range(V.shape[1]):
            if torch.sum(S_norm[:i+1]) < 0.5 and S_norm[i] > 0.1:
                polysemantic.append(i)
        return polysemantic if len(polysemantic) >= threshold else []



# Test this.
import copy
import torch
import torch.nn as  nn
from .base_interpret import BaseInterpretabilityTool
class FeatureViz(BaseInterpretabilityTool):
    """Visulizing what neurons or features are sensitive to in a model."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model
    
    def activate_neuron(self, input_ids, layer_name, neuron_idx, top_k=10):
        """Find inputs that most activate a given neuron in a given layer of the model."""
        max_activations = []
        for ids in input_ids:
            _, cache = self.model(ids)
            activations = cache[layer_name][:, :, neuron_idx]
            max_act, max_idx = torch.max(activations, dim=1)
            max_activations.append(max_act)
        
        return sorted(max_activations, reverse=True)[:top_k]
    
    def neuron_interpolation(self, input_ids, layer_name, neuron_idx1, neuron_idx2, num_interpolations=10, max_new_tokens=10):
        """Interpolate between two neurons to understand their sematic space."""
        alphas = torch.linspace(0, 1, num_interpolations)
        interpolations = []

        for al in alphas:
            model_copy = copy.deepcopy(self.model) # we can use a context manager as well.
            layer = dict(model_copy.named_modules())[layer_name]

            if isinstance(layer, nn.Linear):
                layer.weight.data[:, neuron_idx1] = (1 - al) * self.model.state_dict()[f"{layer_name}.weight"][:, neuron_idx1] + al * self.model.state_dict()[f"{layer_name}.weight"][:, neuron_idx2]
        
            output_ids = model_copy(input_ids)
            interpolations.append(output_ids)
        
        return interpolations


# Update - It works. Tests have been done.
import torch
from .base_interpret import BaseInterpretabilityTool

class AttentionTools(BaseInterpretabilityTool):
    """Understanding attention patterns by analyzing, visualizing and manipulating it."""

    def __init__(self, model) -> None:
        super().__init__(model)
        self.model = model

    #Works
    def attention_patterns(self, input_ids, layer_idx, head_idx):
        """Extract the attention patterns of a specific head in a specific layer of a model."""
        _, cache = self.model(input_ids)
        q = self.model.model_attrs.get_q(layer_idx).weight[head_idx]
        k = self.model.model_attrs.get_k(layer_idx).weight[head_idx]

        # In the mixed models, there seems to be some problem. Q,K, V projections are not matching :)
        # q1 = cache[f"h.{layer_idx}.attn.c_attn.hook_resid_post"][head_idx]
        # k1 = cache[f"h.{layer_idx}.attn.c_attn.hook_resid_post"][head_idx]
        

        return torch.softmax(q @ k / q.shape[-1]**0.5 , dim=-1)
    
    #Works
    def top_attended_ids(self, input_ids, layer_idx, head_idx, num_tokens=2):
        """Extract the top-k attended ids of a specific head in a specific layer of a model."""
        attention = self.attention_patterns(input_ids, layer_idx, head_idx)
        _, top_k_indices = torch.topk(attention, num_tokens, dim=-1)
        return [input_ids[0, idx] for idx in [top_k_indices.item()]]
    
    #Works.
    def ablate_head(self,layer_idx, head_idx):
        """Ablate a specific head in a specific layer of a model."""
        
        W_q = self.model.model_attrs.get_q(layer_idx).weight.clone()
        W_k = self.model.model_attrs.get_k(layer_idx).weight.clone()
        W_v = self.model.model_attrs.get_v(layer_idx).weight.clone()
        
        # Head clone is basically these three sub-networks arranged vertically :
        W_q[:, head_idx] = 0.0
        W_k[:, head_idx] = 0.0
        W_v[:, head_idx] = 0.0
        
        head_clone = torch.cat([W_q, W_k, W_v], dim=0)

        # W_k = 0 ; W_q = 0 ; W_v = 0
        return head_clone

    #Works
    def merge_heads(self, layer_idx, head_idx1, head_idx2, alpha=0.5):
        """Merge two heads in a specific layer of a model. This is really helpful in understanding redundancy in attention heads."""

        # For gpt neo, it is attn.attention.(q_proj, k_proj, v_proj, out_proj).weight, 

        # This is the new way :)
        W_q = self.model.model_attrs.get_q(layer_idx).weight 
        W_k = self.model.model_attrs.get_k(layer_idx).weight
        W_v = self.model.model_attrs.get_v(layer_idx).weight
        W_o = self.model.model_attrs.get_o(layer_idx).weight

        # In Pyotrch, we can't modify tensors that are part of the computations graph in-place if they require grad.
        W_q_copy = W_q.clone()
        W_k_copy = W_k.clone()
        W_v_copy = W_v.clone()
        W_o_copy = W_o.clone()


        W_q_copy[:, head_idx1] = alpha * W_q[:, head_idx1] + (1 - alpha) * W_q[:, head_idx2]
        W_k_copy[:, head_idx1] = alpha * W_k[:, head_idx1] + (1 - alpha) * W_k[:, head_idx2]
        W_v_copy[:, head_idx1] = alpha * W_v[:, head_idx1] + (1 - alpha) * W_v[:, head_idx2]
        W_o_copy[:, head_idx1] = alpha * W_o[:, head_idx1] + (1 - alpha) * W_o[:, head_idx2]

        W_q_copy[:, head_idx2] = 0
        W_k_copy[:, head_idx2] = 0
        W_v_copy[:, head_idx2] = 0
        W_o_copy[:, head_idx2] = 0

        return W_q_copy, W_k_copy, W_v_copy, W_o_copy


import torch
from .base_interpret import BaseInterpretabilityTool

class LogitAttributor(BaseInterpretabilityTool):
    """Attribution of the logit to the input tokens."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model
    
    # Works as of now, but need more testing. GPT test passed.
    def logit_attribution(self, input_ids, target_idx):
        """Computes the attribution of the target logit to the input tokens. Returns the attributions."""
        _, cache = self.model(input_ids)
        logits = cache["logits"]
        target_logit = logits[0, target_idx]
        target_logit.requires_grad = True

        attributions = {}

        # print(list(self.model.named_parameters()))
        # input()

        for name, param in self.model.named_parameters():
            # We will make some arrangments for the final layers as well in the ModelAttributes class.
            if "W_u" in name or "ln_f" in name or ".norm." in name:  # Focus on final layer. 
                param.requires_grad = True
                attributions[name ] = 0.0
        
        target_logit.backward(retain_graph=True)

        for name, param in self.model.named_parameters():
            if "W_u" in name or "ln_f" in name or ".norm." in name:
                if param.grad is not None:
                    attributions[name] += (param * param.grad).sum().item()
                if param.grad is not None:
                    param.grad.zero_()  # Clear the gradients for the next computation

        return attributions
    
    # Works as of now, but need more testing. GPT test passed.
    def track_token_circulation(self, input_ids, target_idx, n_layers=2):
        """Tracks the token circulation through the layers. Returns the direct moves and the total moves."""
        total_moves = 0
        direct_moves = {i: 0 for i in range(n_layers)}
        
        _, cache = self.model(input_ids)
        block_type = self.model.model_attrs.get_block_type()
        
        for i in range(1, n_layers):
            prev_resid = cache[f"{block_type}.{i-1}.hook_resid_post"][0, target_idx]
            curr_attn = cache[f"{block_type}.{i}.attn.hook_result_post"][0, target_idx]
            curr_mlp = cache[f"{block_type}.{i}.mlp.hook_result_post"][0, target_idx]
            curr_resid = cache[f"{block_type}.{i}.hook_resid_post"][0, target_idx]
            
            move_dist = torch.norm(curr_resid - prev_resid)
            total_moves += move_dist.item()
            
            direct_moves[i] += torch.norm(curr_attn) + torch.norm(curr_mlp)
        
        # Normalize direct moves
        for i in direct_moves:
            direct_moves[i] /= total_moves
        
        return direct_moves



# Update, this is working perfectly.
import torch
from .base_interpret import BaseInterpretabilityTool

class AttentionComposer(BaseInterpretabilityTool):
    """Class to interpret attention heads."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model

    #Works
    def head_composition(self, head1, head2, alpha=0.5):
        """Composes two heads."""
        layer1, idx1 = head1
        layer2, idx2 = head2
        
        Q1 = self.model.model_attrs.get_q(layer1).weight[:, idx1]
        K1 = self.model.model_attrs.get_k(layer1).weight[:, idx1]
        V1 = self.model.model_attrs.get_v(layer1).weight[:, idx1]
        O1 = self.model.model_attrs.get_o(layer1).weight[:, idx1]

        Q2 = self.model.model_attrs.get_q(layer2).weight[:, idx2]
        K2 = self.model.model_attrs.get_k(layer2).weight[:, idx2]
        V2 = self.model.model_attrs.get_v(layer2).weight[:, idx2]
        O2 = self.model.model_attrs.get_o(layer2).weight[:, idx2]
        

        # Composite head: attend to what both heads attend to
        Q = (Q1 + Q2) / 2
        K = (K1 + K2) / 2
        V = (V1 + V2) / 2
        O = alpha * O1 + (1 - alpha) * O2 # This is basically "blending" the output. We saw this in Computer Vision!

        return (Q, K, V, O)

    #Works
    def apply_composed_head(self, input_ids, Q, K, V, O):
        """Applies the composed head to the input."""
        _, cache = self.model(input_ids)
        query = cache["resid_pre"] @ Q
        key = cache["resid_pre"] @ K
        value = cache["resid_pre"] @ V
        
        # attn = torch.softmax(query @ key.T / query.shape[-1]**0.5, dim=-1)
        attn = torch.einsum("...qd, ...kd -> qk", query, key) / query.shape[-1]**0.5
        
        result = (attn @ value) @ O.T
        
        return result

    # Works. Works.
    def attention_path_patching(self, input_ids, src_head, dest_head):
        """Patches the attention path from a source head to a destination head."""
        src_layer, src_idx = src_head
        dest_layer, dest_idx = dest_head
        
        def patch_fn(module, input, output):
            src_q = self.model.model_attrs.get_q(src_layer).weight[:, src_idx]
            src_k = self.model.model_attrs.get_k(src_layer).weight[:, src_idx]
            src_v = self.model.model_attrs.get_v(src_layer).weight[:, src_idx]
            if input:
                q = input[0] @ src_q
                k = input[0] @ src_k
                v = input[0] @ src_v
                
                # attn = torch.softmax(q @ k.T / q.shape[-1]**0.5, dim=-1)
                attn = torch.einsum("...qd, ...kd -> qk", q, k) / q.shape[-1]**0.5
                patched = (attn @ v).unsqueeze(-1)
                
                output[0][:, :, dest_idx] = patched.squeeze(-1)  # Patch the output tensor. output[0] is the output tensor.
                return output
        
        handle = self.model.model_attrs.get_attn(dest_layer).register_forward_hook(patch_fn) # this means input and output path will be there.
        result, _ = self.model(input_ids)
        handle.remove()
        return result


# Update: Works but needs to be tested.
import torch
from .base_interpret import BaseInterpretabilityTool

class SuperpositionDisentangler(BaseInterpretabilityTool):
    """Disentangles the superposition of the activations."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model

    # Works, but needs to be tested.
    def compute_activation_covariance(self, ids_list, layer):
        """Computes the covariance matrix of the activations."""
        activations = []
        for input_ids in ids_list:
            _, cache = self.model(input_ids)
            activations.append(cache[layer].squeeze())
        activations = torch.stack(activations)
        return torch.cov(activations.transpose(-2, -1)[:,:-1].reshape(activations.shape[1], -1)) # I'm not sure whether this is correct or not.

    # Works, but needs to be tested.
    def disentangle_superposition(self, cov_matrix, n_concepts=10):
        """Disentangles the superposition of the activations."""
        eigenvalues, eigenvectors = torch.linalg.eigh(cov_matrix)
        idx = torch.argsort(eigenvalues, descending=True)
        return eigenvectors[:, idx[:n_concepts]]

    # The rest need to be udnerstood and tested.
    def project_activations(self, activations, concepts):
        """Projects the activations onto the concepts."""
        return activations @ concepts

    def reconstruct_activations(self, projections, concepts):
        """Reconstructs the activations from the projections."""
        return projections @ concepts.T

    def interpolate_concepts(self, concept1, concept2, alpha):
        """Interpolates between two concepts."""
        return (1 - alpha) * concept1 + alpha * concept2


# Layer attribution and layer attribution. Till now, working for GPT2.
from .base_interpret import BaseInterpretabilityTool

class LogitLens(BaseInterpretabilityTool):
    """Analyzes the logit lens of the model."""
    def __init__(self, model) -> None:
        super().__init__(model)
        self.model = model 
    
    def layer_attributions(self, input_ids, target_idx):
        """Computes the attributions of the layers."""
        _, cache = self.model(input_ids)
        logits = []
        block_type = self.model.model_attrs.get_block_type()

        for i in range(len(self.model.model_attrs.get_block())):
            block_logit = cache[f"{block_type}.{i}.hook_resid_post"] @ self.model.model_attrs.get_lin_ff().weight
            logits.append(block_logit)

        final_logit = cache[f"{block_type}.{len(self.model.model_attrs.get_block())-1}.hook_resid_post"] @ self.model.model_attrs.get_lin_ff().weight
        logits.append(final_logit)

        attributions = [logit[0, target_idx].item() for logit in logits]
        return attributions

    # Same issue. Need to see what are W_u and such.
    def logit_lens(self, input_ids, target_idx, layer_idx):
        """Computes the logit lens of the model."""
        #compares what a layer things the logit should be vs the actual impact.
        _, cache = self.model(input_ids)
        block_type = self.model.model_attrs.get_block_type()
        #direct logit
        direct_logit = cache["logits"]

        # indirect logit
        residual_stack = cache[f"{block_type}.{layer_idx}.hook_resid_post"]
        # print(residual_stack.shape)
        # input()
        # res = residual_stack.clone()
        # print(len(self.model.model_attrs.get_block()))

        # Models that are not working are llama, gemma, phi3(position_ids issue), and bloom(more args in block forward call)
        for i in range(layer_idx+1, len(self.model.model_attrs.get_block()) -1):
            
            block = self.model.model_attrs.get_block()[i]
            
            if type(residual_stack) == tuple:
                # This too is a hacky fix. There are some problems with the implementation.
                continue
            
            #Here, bloom breaks as it requires an alibi and attention mask parameter. Will see, till now bloom is dropped.
            import torch
            if self.model.name in ["llama", "gemma", "phi3"]: 
                # This is such a hacky fix. I tried understanding the hf code, but it seems same to me. The position_ids should be optionalt, but somehow it is necessary for llama, gemma, and phi3.
                position_ids = torch.arange(0, input_ids.shape[1], device=input_ids.device).unsqueeze(0)
                residual_stack = block(residual_stack, position_ids=position_ids) # this is the bitch line.
            else:
                residual_stack = block(residual_stack)

        # print(residual_stack)
        # input()
        indirect_logit = residual_stack[0] @ self.model.model_attrs.get_lin_ff().weight

        return direct_logit[0, target_idx].item(), indirect_logit[0, target_idx].item()


# Update: Works as intended.
import torch
from .base_interpret import BaseInterpretabilityTool 

class KnowledgeProber(BaseInterpretabilityTool):
    """Probes the model for knowledge."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model
    
    #Works 
    def polysemantic_score(self, layer_idx, k_neighbors=5):
        """Computes the polysemantic score of the neurons in the layer."""
        # W = self.model.model.h[layer_idx].mlp.c_fc.weight        # This is for gpt style models.
        W = self.model.model_attrs.get_mlp_in(layer_idx).weight #Check whether this is in weight or out weight.
        
        norms = torch.norm(W, dim=0)
        W_norm = W / norms
        
        similarity = W_norm.T @ W_norm
        _, topk = torch.topk(similarity, k_neighbors + 1, dim=1)
        
        scores = []
        for i in range(W.shape[1]):
            neighbors = W_norm[:, topk[i, 1:]]  # Exclude self
            centrality = torch.norm(neighbors.mean(dim=1))
            dispersion = torch.norm(neighbors - neighbors.mean(dim=1, keepdim=True), dim=0).mean()
            scores.append(centrality / dispersion)
        
        return torch.tensor(scores)
    
    # Works.
    def knowledge_probe(self, ids, concept_neuron):
        """Probes the model for knowledge."""
        true_scores = []
        false_scores = []
        
        layer_idx, neuron_idx = concept_neuron
        
        for true_ids, false_ids in ids:
            _, true_cache = self.model(true_ids)  # This is a caveat.
            _, false_cache = self.model(false_ids)

            block_type = self.model.model_attrs.get_block_type() 
            true_act = true_cache[f"{block_type}.{layer_idx}.mlp.hook_result_post"][0, neuron_idx]
            false_act = false_cache[f"{block_type}.{layer_idx}.mlp.hook_result_post"][0, neuron_idx]
            
            true_scores.append(true_act.item())
            false_scores.append(false_act.item())
        
        return true_scores, false_scores



import torch 
from .base_interpret import BaseInterpretabilityTool
class ResidualDecomposer(BaseInterpretabilityTool):
    """Decomposes the residual stream into interpretable components from MLP and attention."""
    def __init__(self, model) -> None:
        super().__init__(model)
        self.model = model 
    
    # Works.
    def mlp_basis(self, layer_idx):
        """Computes the basis of the MLP layer."""
        W_in = self.model.model_attrs.get_mlp_in(layer_idx).weight
        W_out = self.model.model_attrs.get_mlp_out(layer_idx).weight
        # print(W_in.shape, W_out.shape)
        U, S, V = torch.svd(W_out @ W_in) 
        # print(U.shape, S.shape, V.shape) 

        d_mlp = max(W_in.shape)  # Let's see.
        # print(U[:, :d_mlp].shape, V[:d_mlp].shape) 
        return U[:, :d_mlp], V[:d_mlp]  # REFACTOR: this is d_mlp

    # Breaks for models that have same Q,K,V projection. 
    def decompose_residual(self, input_ids, layer_idx, top_k=1):
        """Decomposes the residual stream into interpretable components from MLP and attention."""
        _, cache = self.model(input_ids)
        # Attention decomposition. It should break here.
        W_q = self.model.model_attrs.get_q(layer_idx).weight
        W_k = self.model.model_attrs.get_k(layer_idx).weight
        W_v = self.model.model_attrs.get_v(layer_idx).weight
        W_o = self.model.model_attrs.get_o(layer_idx).weight

        # print(W_q.shape, W_k.shape, W_v.shape, W_o.shape)
        block_type = self.model.model_attrs.get_block_type()
        
        resid = cache[f"{block_type}.{layer_idx}.hook_resid_post"]
        # print(resid.shape)

        U_mlp, V_mlp = self.mlp_basis(layer_idx)
        mlp_comps = (resid @ V_mlp) @ U_mlp  # let it break here, for GPT models till we get a fix.
        
        scores = torch.einsum("...qd, ...kd -> qk", W_q, W_k) / W_q.shape[-1]**0.5
        
        attn_pattern = torch.softmax(scores, dim=-1)
        # attn_pattern = torch.softmax(Q @ K.transpose(-2, -1) / Q.shape[-1]**0.5, dim=-1)
        # print(attn_pattern.shape, W_v.shape, W_o.shape)

        attn_comps = (attn_pattern @ W_v) @ W_o

        # Top components
        mlp_norms = torch.norm(mlp_comps, dim=0)
        attn_norms = torch.norm(attn_comps.reshape(attn_comps.shape[0], -1), dim=0)

        top_mlp = torch.topk(mlp_norms, top_k)
        top_attn = torch.topk(attn_norms, top_k)

        return {
            "mlp": {"components": mlp_comps, "top_k": (top_mlp.values, top_mlp.indices)},
            "attn": {"components": attn_comps, "top_k": (top_attn.values, top_attn.indices)},
        }


import torch
from .base_interpret import BaseInterpretabilityTool
class CausalTracer(BaseInterpretabilityTool):
    """ locating and Editing Factored Cognition."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model

    # Works on GPT2 style models.
    def trace_token_probability(self, input_ids, target_idx):
        """Traces the probability of the target token through the layers."""
        _, cache = self.model(input_ids)
        probs = []

        for layer_idx in range(len(self.model.model.h)):  # h is for gpt style models.
            resid = cache[f"blocks.{layer_idx}.hook_resid_post"].clone().requires_grad_(True)
            logits = cache['logits'].clone().requires_grad_(True)

            prob = torch.softmax(logits, dim=-1)[0, target_idx]
            probs.append(prob.item())

            # Compute gradient for causal tracing
            prob.backward()
            grad = resid.grad.norm(dim=-1).squeeze().cpu()
            yield prob.item(), grad

        probs.append(torch.softmax(logits, dim=-1)[0, target_idx].item())
        yield probs[-1], None

    # Works on GPT2 style models.
    def intervene_on_neuron(self, input_ids, layer_idx, neuron_idx, scale=0):
        """Intervenes on a neuron in the model."""
        def hook_fn(module, input, output):
            output[:, neuron_idx] = scale * output[:, neuron_idx]
            return output

        handle = self.model.model_attrs.get_mlp(layer_idx).register_forward_hook(hook_fn)
        output = self.model.model(input_ids)
        handle.remove()
        return output



# This, introduced by EleutherAI, is a more robust version of LogitLens.
from contextlib import contextmanager
from .base_interpret import BaseInterpretabilityTool
# This is a really hacky class as of now, so test it more, and if it works, then include it.
class ModelSurgery(BaseInterpretabilityTool):
    """Performs surgery on the model. Replaces, deletes, or permutes layers(in context)."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model

    def get_model(self):
        return self.model
    
    @contextmanager
    def replace_layers(self, indices, replacements):
        """Replaces the specified layers with the given replacements(in context)."""
        # Access layers
        layers = self.model.model_attrs.get_block()
        # Replace specified layers with new ones
        for index, replacement in zip(indices, replacements):
            layers[index] = replacement
        # Update the model's layers
        self.model.model_attrs.set_block(layers)
        try:
            yield self.model
        finally:
            # Restore the original layers
            for index in indices:
                layers[index] = None
            self.model.model_attrs.set_block(layers)

    @contextmanager
    def delete_layers(self, indices):
        """Deletes the specified layers(in context)."""
        # Access layers
        layers = self.model.model_attrs.get_block()
        # Delete specified layers
        for index in sorted(indices, reverse=True):
            del layers[index]
        # Update the model's layers
        self.model.model_attrs.set_block(layers)
        try:
            yield self.model
        finally:
            # Restore the deleted layers
            for index in sorted(indices):
                layers.insert(index, None)
            self.model.model_attrs.set_block(layers)

    @contextmanager
    def permute_layers(self, indices):
        """Permutes the layers based on the given indices(in context)."""
        # Access layers
        layers = self.model.model_attrs.get_block()
        # Permute the layers based on the given indices
        permuted_layers = [layers[i] for i in indices]
        # Update the model's layers
        self.model.model_attrs.set_block(layers)
        try:
            yield self.model
        finally:
            # Restore the original layer order
            self.model.model_attrs.set_block(layers)


import torch
from .base_interpret import BaseInterpretabilityTool

class WriteReadAnalyzer(BaseInterpretabilityTool):
    """Analyzes the read and write heads of the model."""
    def __init__(self, model):
        super().__init__(model)
        self.model = model
    
    #Works
    def identify_write_heads(self, layer_idx, threshold=0.8):
        """Identifies the write heads of the layer. Returns the indices of the heads."""
        W_v = self.model.model_attrs.get_v(layer_idx).weight
        W_o = self.model.model_attrs.get_o(layer_idx).weight

        write_strengths = torch.norm(W_o, dim=0)
        top_k = int(threshold * W_o.shape[1])
        _, indices = torch.topk(write_strengths, top_k)
        
        return [(layer_idx, idx.item()) for idx in indices]
    
    # Works.
    def identify_read_heads(self, layer_idx, dim_idx, threshold=0.8):
        """Identifies the read heads of the layer. Returns the indices of the heads."""
        W_q = self.model.model_attrs.get_q(layer_idx).weight
        W_k = self.model.model_attrs.get_k(layer_idx).weight

        q_aligned = torch.cosine_similarity(W_q[:, dim_idx], W_k.transpose(-2, -1))  # Check if this is correct.
        top_k = int(threshold * q_aligned.shape[0])
        _, indices = torch.topk(q_aligned, top_k)
        
        return [(layer_idx, idx.item()) for idx in indices]
    
    # Works.
    def trace_information_flow(self, input_ids, src_token, dst_token, n_layers=2):
        """Traces the information flow from the source token to the destination token. Returns the flow at each layer."""
        src_idx = input_ids[0].tolist().index(src_token)
        dst_idx = input_ids[0].tolist().index(dst_token)
        
        flow = []
        _, cache = self.model(input_ids)
        
        for i in range(n_layers):
            W_q = self.model.model_attrs.get_q().weight
            W_k = self.model.model_attrs.get_k().weight
            W_v = self.model.model_attrs.get_v().weight
            
            Q = cache[f"blocks.{i}.hook_resid_pre"] @ W_q
            K = cache[f"blocks.{i}.hook_resid_pre"] @ W_k
            V = cache[f"blocks.{i}.hook_resid_pre"] @ W_v
            
            attn = torch.softmax(Q @ K.transpose(-2, -1) / Q.shape[-1]**0.5, dim=-1)
            flow.append(attn[dst_idx, src_idx])
        
        return flow



import json
import time
import torch
import statistics as stats  # see if this import is there or something else.
import hashlib
import inspect
from datetime import datetime
from collections import defaultdict
from functools import wraps
from arrakis.src.core_arrakis.attention_tools import AttentionTools
from arrakis.src.core_arrakis.logit_attribution import LogitAttributor
from arrakis.src.core_arrakis.knowledge_prober import KnowledgeProber
from arrakis.src.core_arrakis.causal_tracing_intervention import CausalTracer
from arrakis.src.core_arrakis.superposition_disentangler import SuperpositionDisentangler
from arrakis.src.core_arrakis.sparsity_analyzer import SparsityAnalyzer
from arrakis.src.core_arrakis.read_write_heads import WriteReadAnalyzer
from arrakis.src.core_arrakis.attention_head_comp import AttentionComposer
from arrakis.src.core_arrakis.kg_extractor import KnowledgeGraphExtractor
from arrakis.src.core_arrakis.logit_lens import LogitLens
from arrakis.src.core_arrakis.residual_decomposer import ResidualDecomposer
from arrakis.src.core_arrakis.residual_tools import ResidualTools
from arrakis.src.core_arrakis.knowledge_prober import KnowledgeProber
from arrakis.src.core_arrakis.model_surgery import ModelSurgery
from arrakis.src.core_arrakis.feature_viz import FeatureViz
"""Base Interpretablity Bench. Keep in mind we are not dealing with tokenization and everything. We need the ids, an we'll work with that."""
class BaseInterpretabilityBench:
    def __init__(self, model, save_dir="experiments"):
        self.model = model
        self.save_dir = save_dir
        self.tools = {
            "attention": AttentionTools(model), 
            "composer": AttentionComposer(model), 
            "logit": LogitAttributor(model), 
            "logit_lens": LogitLens(model),
            "knowledge": KnowledgeProber(model), 
            "kg_extractor": KnowledgeGraphExtractor(model),
            "causal": CausalTracer(model), 
            "decomposer": ResidualDecomposer(model),
            "superposition": SuperpositionDisentangler(model), 
            "sparsity": SparsityAnalyzer(model), 
            "write_read": WriteReadAnalyzer(model),
            "residual": ResidualTools(model),
            "k_prober": KnowledgeProber(model),
            "surgery": ModelSurgery(model),
            "feature_viz": FeatureViz(model)

            # Add more tools here
        }

        self._experiment_versions = {}
        self._experiments = {}
        self._plotting_lib = None

    def set_plotting_lib(self, lib):
        self._plotting_lib = lib

    def plot_results(self, *plots, input_ids, **kwargs):
        def decorator(func) :
            @wraps(func)
            def wrapper(*args, **kwargs):
                if self._plotting_lib is None:
                    raise ValueError("Plotting library not set. Use set_plotting_lib() first.")
                
                figs = []
                _, cache = self.model(input_ids)
                for plot in plots:
                    fig = self._plotting_lib.plot(
                        self._plotting_lib,
                        cache = cache,
                        plot_type=plot.plot_type,
                        data_keys=plot.data_keys,
                        title=plot.title,
                        **kwargs
                    )
                    figs.append(fig)
                
                return figs
            return wrapper
        return decorator


    def log_experiment(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            experiment_name = func.__name__
            print(f"Running experiment: {experiment_name}")
            result = func(*args, **kwargs)
            # print("Result: ", result)
            exp_data = {
                "args": args,
                "kwargs": kwargs,
                "result": result,
                "timestamp": datetime.now().isoformat(),
                "source": inspect.getsource(func)
            }

            # print(exp_data)

            hasher = hashlib.sha256()
            hasher.update(json.dumps(exp_data, default=self.json_encoder).encode())
            version_hash = hasher.hexdigest()[:8]  # First 8 characters of the hash
            
            if func.__name__ not in self._experiment_versions:
                self._experiment_versions[func.__name__] = []
            self._experiment_versions[func.__name__].append(version_hash)
            
            self._experiments[f"{func.__name__}_v{version_hash}"] = exp_data

            # Save results
            save_path = f"{self.save_dir}/{experiment_name}.json"
            with open(save_path, "w") as f:
                json.dump(exp_data, f, indent=4, default=self.json_encoder)
            
            print(f"Results saved to {save_path}")
            return result
        return wrapper

    def list_versions(self, experiment_name):
        return self._experiment_versions.get(experiment_name, [])
    
    def get_version(self, experiment_name, version_hash):
        key = f"{experiment_name}_v{version_hash}"
        return self._experiments.get(key, None)
    
    def test_hypothesis(self, null_func, alt_func, n_samples=100, alpha=0.05):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                result = func(*args, **kwargs)

                # Generate samples under null and alternate hypothesis
                null_samples = [null_func(*args, **kwargs) for _ in range(n_samples)]
                alt_samples = [alt_func(*args, **kwargs) for _ in range(n_samples)]

                t_stat, p_value = stats.ttest_ind(null_samples, alt_samples)

                result["hypothesis_test"] = {
                    "t_stat": t_stat,
                    "p_value": p_value,
                    "reject_null": p_value < alpha
                }
                return result 
            return wrapper
        return decorator
    
    def profile_model(self, granularity="layer"):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                model = self.model
                timings = defaultdict(float)
                activation_sizes = defaultdict(list)

                def layer_hook(name):
                    def hook(module, input, output):
                        t0 = time.time()
                        yield 
                        timings[name] += time.time() - t0
                        activation_sizes[name].append(output.numel() * output.element_size())
                    return hook
                
                hooks = []
                for name, module in model.named_modules():
                    if granularity == "layer" and "layers." in name:
                        hooks.append(module.register_forward_hook(layer_hook(name), prepend=True))
                    elif granularity == "component" and any(x in name for x in ["attention", "mlp", "layer_norm"]):
                        hooks.append(module.register_forward_hook(layer_hook(name), prepend=True))

                result = func(model, *args[1:], **kwargs)

                for hook in hooks:
                    hook.remove()

                total_time = sum(timings.values())
                total_memory = sum(max(sizes) for sizes in activation_sizes.values())

                result["profiling"] = {
                    "time_percentages": {k: v / total_time for k, v in timings.items()},
                    "memory_percentages": {k: max(v) / total_memory for k, v in activation_sizes.items()},
                    "total_time": total_time,
                    "total_memory": total_memory
                }
                return result
            return wrapper
        return decorator
    
    def use_tools(self, *tools):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                required_tools = [self.tools[tool] for tool in tools]
                return func(*args, **kwargs, **{tool: tool_instance for tool, tool_instance in zip(tools, required_tools)})
            return wrapper
        return decorator


    @staticmethod
    def json_encoder(obj):
        if isinstance(obj, torch.Tensor):
            return obj.tolist()
        return str(obj)


import torch
from .base_interpret import BaseInterpretabilityTool

class ResidualTools(BaseInterpretabilityTool):
    """The residual stream in transformers carries information throughout the network."""
    def __init__(self, model) -> None:
        super().__init__(model)
        self.model = model 

    #Works.
    def residual_decomposition(self, input_ids, layer_idx):
        """Break down the residual stream to see where info comes from."""
        # See this whether the names are correct or not.
        _, cache = self.model(input_ids)
        block_type = self.model.model_attrs.get_block_type() 
        residual = cache[f"{block_type}.{layer_idx}.hook_resid_pre"]
        attn_out = cache[f"{block_type}.{layer_idx}.attn.hook_result"]
        mlp_out  = cache[f"{block_type}.{layer_idx}.mlp.hook_result"]

        return {
            "from_prev": residual-attn_out-mlp_out,
            "from_attn": attn_out,
            "from_mlp": mlp_out
        }
    
    #Works.
    def residual_movement(self, input_ids, layer_idx1, layer_idx2):
        """ Measure how much the residual stream changes between layers."""
        _, cache = self.model(input_ids)
        block_type = self.model.model_attrs.get_block_type()
        residual1 = cache[f"{block_type}.{layer_idx1}.hook_resid_post"]
        residual2 = cache[f"{block_type}.{layer_idx2}.hook_resid_pre"]
        
        movement = torch.norm(residual2 - residual1, dim=-1) / torch.norm(residual1, dim=-1)
        return movement.mean().item()
    
    # Logits part is not working.
    def feature_attribution(self, input_ids, target_ids):
        """Attribute the final prediction back through the residual stream."""
        _, cache = self.model(input_ids)
        block_type = self.model.model_attrs.get_block_type()
        logits = cache["logits"]  # BREAK.
        target_logits = logits[0, target_ids]

        residuals = [cache[f"{block_type}.{i}.hook_resid_post"] for i in range(len(self.model.model_attrs.get_block()))]
        attributions = []

        for resid in residuals:
            resid.requires_grad = True
            logits = self.model.model_attrs.get_lin_ff()(resid) # BREAK.
            # logits = self.model.model_attrs.get_embed()(logits) # BREAK.

            target_logit = logits[0, target_ids]
            target_logit.mean().backward() # This fixes it, but I'm not sure how correct it is.

            attributions.append(torch.norm(resid.grad, dim=-1))
        
        return attributions


# Have to incorporate it others.
import numpy as np

class FactoredMatrix:
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def eigenvalues(self):
        # Calculate the eigenvalues of the product of left and right matrices
        return np.linalg.eigvals(np.matmul(self.left, self.right))

    def norm(self, order=2):
        # Calculate the matrix norm of the product of left and right matrices
        return np.linalg.norm(np.matmul(self.left, self.right), ord=order)

    def svd(self):
        # Calculate the SVD of the product of left and right matrices
        return np.linalg.svd(np.matmul(self.left, self.right))


from transformers import PreTrainedModel, PretrainedConfig
from transformers import GPT2Model, GPTNeoModel, GPTNeoXModel, LlamaModel, GemmaModel, Phi3Model, T5Model, Qwen2Model, BloomModel, MistralModel, StableLmModel, MixtralModel
from transformers import GPT2Config, GPTNeoConfig, GPTNeoXConfig, LlamaConfig, GemmaConfig, Phi3Config, T5Config, Qwen2Config, BloomConfig, MistralConfig, StableLmConfig, MixtralConfig

import torch 
import torch.nn as nn
import logging
import pathlib 
"""Helper class to register hooks on a model and get activations."""

#TODO: Can't seem to figure out how there is an empty entry in the activations.
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG, filename=pathlib.Path.joinpath(pathlib.Path.cwd(), "logs", "arrakis_model.log"), filemode="w", format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

H_MODELS = ["gpt2", "gpt-neo"]

LAYER_MODELS = ["llama", "gemma", "phi3", "t5", "qwen2", "mistral", "stable-lm", "gpt-neox"]

DIFFERENT_KINDA = ["bloom"]

MODELS_THAT_ARE_SUPPORTED = [
    "gpt2",
    "gpt-neo",
    "gpt-neox",
    "llama",
    "gemma",
    "phi3",
    "qwen2",
    "mistral",
    "stable-lm",

]

MODEL_CLASS_MAPPING = {
    "gpt2": GPT2Model,
    "gpt-neo": GPTNeoModel,
    "gpt-neox": GPTNeoXModel,
    "llama": LlamaModel,
    "gemma": GemmaModel,
    "phi3": Phi3Model,
    "t5": T5Model,
    "qwen2": Qwen2Model,
    "mistral": MistralModel,
    "stable-lm": StableLmModel,
}

MODEL_CONFIG_MAPPING = {
    "gpt2": GPT2Config,
    "gpt-neo": GPTNeoConfig,
    "gpt-neox": GPTNeoXConfig,
    "llama": LlamaConfig,
    "gemma": GemmaConfig,
    "phi3": Phi3Config, 
    "t5": T5Config,
    "qwen2": Qwen2Config,
    "mistral": MistralConfig,
    "stable-lm": StableLmConfig,

}

class ActivationCache:
    """Registers hooks on a model and stores the activations."""
    def __init__(self, model):
        self.model = model
        self.activations = {}
        self.hooks = []
    
    def register_hooks(self):
        """Registers hooks on the model."""
        def hook_fn(name, is_forward=True):
            def register_hook(module, input, output=None):
                # Only works with register_forward_hook(needs output), not pre_hook.
                if is_forward:
                    self.activations[name] = output[0].detach().cpu()
                else:
                    # See this hacky fix. It is working, but idk why. I'll look into it later.
                    if len(input) >= 1:
                        if type(input) == tuple:
                            # print(input)
                            self.activations[name] = input[0].detach().cpu()
                    else:
                        self.activations[name] = torch.tensor(input).detach().cpu()
            return register_hook

        logger.debug(f"Registering hooks for {self.model.name}")
        for name, submodule in self.model.model.named_modules():
            if hasattr(submodule, "register_forward_hook"):
                if name=="": # This is the empty entry that I was talking about. It is the whole model.
                    self.hooks.append(submodule.register_forward_hook(hook_fn("root")))
                else:
                    self.hooks.append(submodule.register_forward_hook(hook_fn(name)))

        # Now, we can just update this as model_attrs takes care for us :)
        logger.debug(f"Registering iterative hooks(or blocks/layers) for {self.model.name}")
        for i, block in enumerate(self.model.model_attrs.get_block()):
            
            block_type = self.model.model_attrs.get_block_type()
            attn_type = self.model.model_attrs.get_attn_type()
            # The ones starting with block are backward hooks. The ones starting with blocks are forward hooks.
            block.register_forward_pre_hook(hook_fn(f"{block_type}.{i}.hook_resid_pre", is_forward=False)) # This line breaks. Fixed
            block.register_forward_hook(hook_fn(f"{block_type}.{i}.hook_resid_post"))
            
            mlp = self.model.model_attrs.get_mlp(i)
            mlp.register_forward_hook(hook_fn(f"{block_type}.{i}.mlp.hook_result_pre", is_forward=False))
            mlp.register_forward_hook(hook_fn(f"{block_type}.{i}.mlp.hook_result_post"))

            
            attn = self.model.model_attrs.get_attn(i)
            attn.register_forward_hook(hook_fn(f"{block_type}.{i}.{attn_type}.hook_result_pre",is_forward=False))
            attn.register_forward_hook(hook_fn(f"{block_type}.{i}.{attn_type}.hook_result_post"))
            
            q = self.model.model_attrs.get_q(i) ; k = self.model.model_attrs.get_k(i) ; v = self.model.model_attrs.get_v(i)
            q_type = self.model.model_attrs.get_q_type() ; k_type = self.model.model_attrs.get_k_type() ; v_type = self.model.model_attrs.get_v_type()
            
            q.register_forward_hook(hook_fn(f"{block_type}.{i}.{attn_type}.{q_type}.hook_result_pre", is_forward=False))
            q.register_forward_hook(hook_fn(f"{block_type}.{i}.{attn_type}.{q_type}.hook_result_post"))

            k.register_forward_hook(hook_fn(f"{block_type}.{i}.{attn_type}.{k_type}.hook_result_pre", is_forward=False))
            k.register_forward_hook(hook_fn(f"{block_type}.{i}.{attn_type}.{k_type}.hook_result_post"))

            v.register_forward_hook(hook_fn(f"{block_type}.{i}.{attn_type}.{v_type}.hook_result_pre", is_forward=False))
            v.register_forward_hook(hook_fn(f"{block_type}.{i}.{attn_type}.{v_type}.hook_result_post"))
            
            

        
        logger.debug(f"Registering hooks for the last module(logits) of the model : {self.model.name}")
        last_module_name = list(self.model.named_modules())[-1][0]
        last_module = dict(self.model.named_modules())[last_module_name]
        last_module.register_forward_hook(hook_fn("logits"))      

def register_hooks(forward_fn):
    """Decorator to register hooks on the model."""
    logger.debug(f"Invoked Activation Cache. Registering hooks for : {forward_fn.__name__}")
    def wrapped_forward(self, *args, **kwargs):
        cache = ActivationCache(self)
        cache.register_hooks()

        output = forward_fn(self, *args, **kwargs)
        activations = cache.activations

        return output, activations
    
    return wrapped_forward

class ModelAttributes:
    """Helper class to get the attributes of the model, their names and attributes."""
    def __init__(self, model):
        self.model = model
        
        # These attributes are used by us in the main library.
        self.block_type = None
        self.attn_type = None
        self.q_type = None
        self.k_type = None
        self.v_type = None
        self.o_type = None
        self.mlp_in_type = None
        self.mlp_out_type = None
        self.lin_ff_type = None # Yes. Used in kg-prober.
        self.embed_type = None # Yes. Used in 


        self._models_exceptions_list = ["get-neox", "phi3", "bloom"]
        self._qkv_exceptions_list = []
        self._o_exceptions_list = []

        self.setup()
    
    def setup(self):
        self.model_str = repr(self.model)
        self.parse(self.model_str)

    def parse(self, model_str):
        """Parses the model string to get the attributes of the model. There are better ways to do this :) """
        # Update; There must be a clever way to do this using ast. I'll look into it.
        # We'll start from the embeddings and go till the output layer.
        # print(model_str)
        if "(wte):" in model_str:
            self.embed_type = "wte"
        elif "(word_embeddings):" in model_str:
            self.embed_type = "word_embeddings"
        elif "(embed_tokens):" in model_str:
            self.embed_type = "embed_tokens"
        elif "(embed_in):" in model_str:
            self.embed_type = "embed_in"
        else:
            self.embed_type = "<-break->"
        
        # First, let's see what are the main blocks called.
        if "(h):" in model_str:
            self.block_type = "h"
        elif "(blocks):" in model_str:
            self.block_type = "blocks"
        elif "(layers):" in model_str:
            self.block_type = "layers"
        else:
            self.block_type = "<-break->"

        block_module = model_str[model_str.index(self.block_type):]
        # print(block_module)
        # Now, let's see what are the main attention blocks called.
        if "(attn):" in block_module:
            self.attn_type = "attn"
        elif "(attention):" in block_module:
            self.attn_type = "attention"
        elif "(self_attn):" in block_module:
            self.attn_type = "self_attn"
        elif "(self_attention):" in block_module:
            self.attn_type = "self_attention"
        else:
            self.attn_type = "<-break->" # attention is now accessed by model.block_type.attn_type

        # Inside attn_type, let's see what are the main attention components called.
        # print(model_str)

        # This is the troublemaker. Here parsing is not accurate.
        attn_module = model_str[model_str.index(self.attn_type):]
        # print(attn_module) 

        if "(c_attn):" in attn_module and "(c_proj):" in attn_module:
            self.q_type = "c_attn" ; self.k_type = "c_attn" ; self.v_type = "c_attn" ; 
            if "(c_proj):" in attn_module:
                self.o_type = "c_proj"

        elif "(q_proj):" in attn_module and "(k_proj):" in attn_module and "(v_proj):" in attn_module:
            self.q_type = "q_proj" ; self.k_type = "k_proj" ; self.v_type = "v_proj"
            if "(out_proj):" in attn_module:
                self.o_type = "out_proj"
            elif "(o_proj):" in attn_module:
                self.o_type = "o_proj"
            else:
                self.o_type = "<-break->"

        else:
            # If we are reaching here, then it means that q_k_v are all bundled. Currently, this happens for get-neox, phi3, and bloom. All are dealt with here.
            if "(query_key_value):" in attn_module:
                self.q_type = "query_key_value"
                self.k_type = "query_key_value"
                self.v_type = "query_key_value"
                if "dense" in attn_module:
                    self.o_type = "dense"
            
            elif "(qkv_proj):" in attn_module:
                self.q_type = "qkv_proj"
                self.k_type = "qkv_proj"
                self.v_type = "qkv_proj"
                if "o_proj" in attn_module:
                    self.o_type = "o_proj"
        
        mlp_module = model_str[model_str.index(self.o_type):]
        # print(mlp_module)
        if "(c_fc):" in mlp_module and "(c_proj):" in mlp_module: # Gpt, GPTNeo
            self.mlp_in_type = "c_fc" ; self.mlp_out_type = "c_proj"
        
        elif "(dense_h_to_4h):" in mlp_module and "(dense_4h_to_h)" in mlp_module:
            self.mlp_in_type = "dense_h_to_4h" ; self.mlp_out_type = "dense_4h_to_h"

        elif "(gate_proj):" in mlp_module and "(down_proj):" in mlp_module: # Gemma, Llama
            self.mlp_in_type = "gate_proj" ; self.mlp_out_type = "down_proj"

        elif "(gate_up_proj):" in mlp_module and "(down_proj):" in mlp_module: # Phi3
            self.mlp_in_type = "gate_up_proj" ; self.mlp_out_type = "down_proj"
        
        elif "(fc_in):" in mlp_module and "(fc_out):" in mlp_module: # Qwen2
            self.mlp_in_type = "fc_in" ; self.mlp_out_type = "fc_out"
        
        lin_ff_module = model_str[model_str.index(self.mlp_out_type):]
        if "(ln_f):" in lin_ff_module:
            self.lin_ff_type = "ln_f"
        elif "(final_layer_norm)" in lin_ff_module:
            self.lin_ff_type = "final_layer_norm"
        elif "(norm):" in lin_ff_module:
            self.lin_ff_type = "norm"
        else:
            self.lin_ff_type = "<-break->"
        # This parser seems to be working with all except get-neox, phi3, and bloom. one reason is that they have combined q_k_v in one layer. output name is also different.
        # Upate: I've added cases that we can cover till now :) We'll deal with the rest later.

    def get_block_type(self):
        """Returns the block type of the model."""
        return self.block_type
    

    def get_attn_type(self):
        """Returns the attention type of the model."""
        return self.attn_type
    
    def get_q_type(self):
        """Returns the query type of the model."""
        return self.q_type
    
    def get_k_type(self):
        """Returns the key type of the model."""
        return self.k_type
    
    def get_v_type(self):
        """Returns the value type of the model."""
        return self.v_type
    
    def get_o_type(self):
        """Returns the output type of the model."""
        return self.o_type
    
    def get_mlp_in_type(self):
        """Returns the input type of the MLP layer."""
        return self.mlp_in_type
    
    def get_mlp_out_type(self):
        """Returns the output type of the MLP layer."""
        return self.mlp_out_type
    
    def get_lin_ff_type(self):
        """Returns the final layer norm type of the model."""
        return self.lin_ff_type
    
    def get_embed_type(self):
        """Returns the embedding type of the model."""
        return self.embed_type
    
    def get_embed(self):
        """Returns the embedding layer of the model."""
        return getattr(self.model, self.embed_type)
    
    def get_lin_ff(self):
        """Returns the final layer norm of the model."""
        return getattr(self.model, self.lin_ff_type)
    
    def get_block(self, layer_idx=None):
        """Returns the block of the model. If layer_idx is provided, returns the block at that index."""
        if layer_idx:
            return getattr(self.model, self.block_type)[layer_idx]  # This might change if we want to make it an attribute inside `HookedModel`
        return getattr(self.model, self.block_type)
    
    def set_block(self, block):
        """Sets the block of the model. Used in model_surgery"""
        setattr(self.model, self.block_type, block)

    
    def get_attn(self, layer_idx):
        """Returns the attention layer of the model."""
        if isinstance(self.get_block(layer_idx), nn.ModuleList):
            return getattr(self.get_block(layer_idx)[0], self.attn_type)
        
        return getattr(self.get_block(layer_idx), self.attn_type) # could be anything.

    def get_mlp(self, layer_idx):
        """Returns the MLP layer of the model."""
        if isinstance(self.get_block(layer_idx), nn.ModuleList):
            return getattr(self.get_block(layer_idx)[0], "mlp")
        return getattr(self.get_block(layer_idx), "mlp")

    def get_mlp_in(self, layer_idx):
        """Returns the input layer of the MLP layer."""
        return getattr(self.get_mlp(layer_idx), self.mlp_in_type)
    
    def get_mlp_out(self, layer_idx):
        """Returns the output layer of the MLP layer."""
        return getattr(self.get_mlp(layer_idx), self.mlp_out_type)

    def get_q(self, layer_idx):
        """Returns the query layer of the attention layer."""
        # If the model is GPTNeo, there is an extra layer of wrapping. GPTNeoAttention -> GPTNeoSelfAttention
        if "GPTNeoModel" in self.model_str:
            return getattr(self.get_attn(layer_idx).attention, self.q_type)
        return getattr(self.get_attn(layer_idx), self.q_type)
    
    def get_k(self, layer_idx):
        """Returns the key layer of the attention layer."""
        if "GPTNeoModel" in self.model_str:
            return getattr(self.get_attn(layer_idx).attention, self.k_type)
        return getattr(self.get_attn(layer_idx), self.k_type)

    def get_v(self, layer_idx):
        """Returns the value layer of the attention layer."""
        if "GPTNeoModel" in self.model_str:
            return getattr(self.get_attn(layer_idx).attention, self.v_type)
        return getattr(self.get_attn(layer_idx), self.v_type)
     
    def get_o(self, layer_idx):
        """Returns the output layer of the attention layer."""
        if "GPTNeoModel" in self.model_str:
            return getattr(self.get_attn(layer_idx).attention, self.o_type)
        return getattr(self.get_attn(layer_idx), self.o_type)
    
    def __repr__(self):
        """Returns the string representation of the model attributes."""
        return f"ModelAttributes(block_type={self.block_type}, attn_type={self.attn_type}, q_type={self.q_type}, k_type={self.k_type}, v_type={self.v_type}, o_type={self.o_type}, mlp_in_type={self.mlp_in_type}, mlp_out_type={self.mlp_out_type})"



class HookedAutoConfig(PretrainedConfig):
    """Config class for the models. This is a wrapper around the HF Pretrained config class."""
    def __init__(self, name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logger.debug(f"Invoked HookedAutoConfig. Creating config for {name}. Args are {args} and kwargs are {kwargs}")
        if name not in MODELS_THAT_ARE_SUPPORTED:
            raise ValueError(f"{name} is not supported. Supported models are {MODELS_THAT_ARE_SUPPORTED}")
        if name not in MODEL_CONFIG_MAPPING:
            raise ValueError(f"{name} is not supported. Supported models are {MODEL_CONFIG_MAPPING.keys()}")
        
        self.name = name
        self.config  = MODEL_CONFIG_MAPPING[name](*args, **kwargs)


class HookedAutoModel(PreTrainedModel):
    """Model class for the models. This is a wrapper around the HF Pretrained model class."""
    def __init__(self, config:HookedAutoConfig, *args, **kwargs):
        super().__init__(config)
        logger.debug(f"Invoked HookedAutoModel. Creating model for {config.name}")
        self.name = config.name  # Already tested what all models are supported.
        self.model = MODEL_CLASS_MAPPING[self.name](config.config, *args, **kwargs)
        self.model_attrs = ModelAttributes(self.model)

    @register_hooks
    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)


# Update: Works. But, we need to test it more.
import torch
from .base_interpret import BaseInterpretabilityTool
class KnowledgeGraphExtractor(BaseInterpretabilityTool):
    """Extracting relational subspaces from Q-K interaction,Projecting token embeddings onto these subspaces, Indentifying types of relations""" 
    def __init__(self, model) -> None:
        super().__init__(model)
        self.model = model 
    
    # Works.
    def extract_relational_subspaces(self, layer_idx, top_k=3):
        """Extracts the top-k relational subspaces from the Q-K interaction matrix of the layer."""
        # attn_block = self.model.model_attrs.get_attn(layer_idx)
        W_q = self.model.model_attrs.get_q(layer_idx).weight
        W_k = self.model.model_attrs.get_k(layer_idx).weight
        # W_k = self.model.blocks[layer_idx].attn.W_k
        
        W_QK = W_q @ W_k.T

        U, S, V = torch.svd(W_QK)  # We can use our FactoredMatrix class here.
        top_subspaces = V[:, :top_k].T 
        return top_subspaces

    # Works
    def project_token_embeddings(self, token_ids, subspaces):
        """Projects the token embeddings onto the given subspaces."""
        embeddings = self.model.model_attrs.get_embed()(token_ids) # It should be self.model.embed afaik.
        return {i: (embeddings @ subspace.T).squeeze() for i, subspace in enumerate(subspaces)}
    
    # Works, but more tests needed.
    def identify_relation_types(self, token_ids, subspaces, threshold=0.3):
        """Identifies the types of relations based on the projected embeddings."""
        projections = self.project_token_embeddings(token_ids, subspaces)
        relations = {i: [] for i in range(len(subspaces))}

        for i, proj in projections.items():
            for j, val in enumerate(proj):
                if (val > threshold).all() :   # See this refactor. if all of the vals is bigger than threshold, we append. 
                    relations[i].append(token_ids[j])
        
        return relations